/***************************************************************/
/*                        FEFNT.C                              */
/*                                                             */
/*  PROGRAM:   Font Editor (nije modul programa)               */
/*  FUNKCIJE:  Funkcije za ispis Font Editor user FNT fontova  */
/*             pod MS DOS-om (screen format)                   */
/*                                                             */
/*  BOOL   MemoryAlloc(HWND);                                  */
/*  BOOL   MemoryLock(void);                                   */
/*  LPSTR  MemoryAdress(int);                                  */
/*  void   MemoryUnlock(void);                                 */
/*  void   MemoryFree(void);                                   */
/*                                                             */
/*  (c) 1994 by Vedran Gaco                                    */
/***************************************************************/

#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if !defined(__NEARPOINTERS) && !defined(__FARPOINTERS)
#if defined(__TINY__) || defined(__SMALL__) || defined(__MEDIUM__)
#define __NEARPOINTERS
#else
#define __FARPOINTERS
#endif
#endif

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

#ifdef __cplusplus
extern "C" {
#endif
void   userinitgraph(int, int);
int    loaduserfont(const char *, const char *, int);
int    deleteuserfont(int);
void   deletealluserfonts(void);
int    selectuserfont(int);
int    usertextheight(void);
int    usertextwidth(const char *);
int    usertextheight(void);
int    userfontbaseline(void);
struct fontheader *getuserfontheader(int);
int    setuserbkmode(int);
int    getuserbkmode(void);
int    userouttext(char *);
int    userouttextxy(int, int, char *);
int    userTextOut(int, int, char *, int);
#ifdef __cplusplus
}
#endif

#define  MAXUSERFONTS  20          /* max broj user fontova */

typedef unsigned char byte;

struct  fontheader {
          char    name[20];        /* ime fonta */
          byte    size;            /* velicina fonta */
          byte    numchars;        /* broj definiranih karaktera u fontu */
          unsigned char defaultch; /* bit ce ispisan umjesto karaktera
                                      koji ne postoje u fontu */
          int     height;          /* visina fonta */
          byte    avwidth;         /* srednja sirina sova */
          byte    baseline;        /* bazna linija */
          byte    blockreserved;   /* koristi se interno (nacin zauzimanja
                                      memorije) */
          byte    *ptrwidth;       /* pointer na niz duzine 256 bajtova koji
                                      sadrzi sirine karaktera (poredanih po
                                      ASCII redoslijedu */
          byte    **ptrbitmap;     /* pointer na niz od 256 pointera koji
                                      sadrze adrese bitmapa */
          byte    *ptrdefined;     /* pointer na niz od numchars bajtova
                                      koji sadrzi sve definirane karaktere
                                      u fontu (redosljed je isti kao u
                                      datoteci tj. ne mora biti ASCII, ali
                                      je pozeljno da bude) */
     };

/* niz pointera na fontove */
static struct fontheader *fontadress[MAXUSERFONTS] =
                            { NULL, NULL, NULL, NULL, NULL,
                              NULL, NULL, NULL, NULL, NULL,
                              NULL, NULL, NULL, NULL, NULL,
                              NULL, NULL, NULL, NULL, NULL };

static int   selectedfont = 0;     /* hanle kurent fonta (inicijalno 0) */


/* Varijable potrebne asemblerskim rutinama za ispis teksta */

static int       CurrGraphDriver = -1;  /* Trenutni graficki driver */
static int       CurrGraphMode = 0;     /* Tekuci graficki mod */
static unsigned  CurrVideoSeg;          /* Segment video adrese */
static unsigned  HercVideoSeg[4]   = { 0xb000, 0xb200, 0xb400, 0xb600 };
static unsigned  ATT400VideoSeg[4] = { 0xb800, 0xba00, 0xbc00, 0xbe00 };
static unsigned  *VideoSegPtr;          /* pointer na HercVideoSeg ili ATT400VideoSeg */
static int       CurrMaxX;              /* max hor rezolucija */
static int       CurrMaxY;              /* max ver rezolucija */
static int       CurrScanBytes;         /* br bajtova u scan liniji */
static byte      TwoBitsPerPixel;       /* br bitova po pixelu */
static byte      GraphVadrRut;          /* Rutina za izrac. video adrese */
static int       FontBitBlt;            /* Rutina za crtanje bitmapa */
                                        /* 0 - default  (putpixel)
                                           1 - direktno u memoriju (CGA, MCGA itd)
                                           2 - pomocu In/Out portova (EGA, VGA)
                                        */
static byte      TextBkMode = OR_PUT;    /* nacin iscrtavanja teksta */

/*********************************************************************/
/*                                                                   */
/*                       userinitgraph                               */
/*                                                                   */
/*   Kako BORLAND-ov BGI nema funkciju getgraphdriver, koja bi       */
/*   vratila trenutno aktivni graficki drajver, potrebno je poslje   */
/*   svake inicijalizacije grafike ili promjene grafickog moda,      */
/*   pozvati ovu funkciju, tj. funkcijama za ispis Font Editor       */
/*   fontova naznaciti aktivni graficki drajver i mod. Mada ovo      */
/*   izgleda komplicirano, ispis teksta je dosta brzi nego da se     */
/*   pri svakom pozivu "userouttext" funkcija poziva getdrivername   */
/*   i onda iz vracenog stringa odredjuje drajver.                   */
/*                                                                   */
/*   Ukratko ova funkcija mora biti pozvana odmah nakon funkcija     */
/*   intigraph i setgraphmode:                                       */
/*                                                                   */
/*   initgraph(&graphdriver, &graphmode, "");                        */
/*   userinitgraph(graphdriver, graphmode);                          */
/*                                                                   */
/*   setgraphmode(graphmode);                                        */
/*   userinitgraph(graphdriver, graphmode);                          */
/*                                                                   */
/*********************************************************************/

void userinitgraph(int gdriver, int gmode)
{
    CurrMaxX = getmaxx();
    CurrMaxY = getmaxy();
    CurrScanBytes = CurrMaxX / 8 + 1;  /* br bajtova u scan liniji */

    CurrGraphMode = gmode;
    switch(CurrGraphDriver = gdriver)
    {
        case CGA:
              CurrVideoSeg = 0xB800;       /* segment video adrese */
              FontBitBlt = 1;              /* rutina za iscrtavanje bitmapa */
              if(gmode == CGAHI)
                 TwoBitsPerPixel = FALSE;  /* 4 boje (2 bita po pikselu) */
              else {
                 TwoBitsPerPixel = TRUE;
                 CurrScanBytes = 80;
              }
              break;

        case MCGA:
              FontBitBlt = 1;
              switch(gmode) {
                 case MCGAHI:
                    TwoBitsPerPixel = FALSE;
                    CurrVideoSeg = 0xA000;
                    GraphVadrRut = 0;
                    break;

                 case MCGAMED:
                    TwoBitsPerPixel = FALSE;
                    CurrVideoSeg = 0xB800;
                    GraphVadrRut = 1;
                    break;

                 default:  /* MCGAC0, MCGAC1, MCGAC2, MCGAC3 */
                    TwoBitsPerPixel = TRUE;
                    CurrScanBytes = 80;
                    CurrVideoSeg = 0xB800;
                    GraphVadrRut = 1;
              }
              break;

        case ATT400:
              FontBitBlt = 1;
              CurrVideoSeg = 0xB800;
              switch(gmode) {
                 case ATT400HI:
                   TwoBitsPerPixel = FALSE;
                   GraphVadrRut = 2;
                   VideoSegPtr = ATT400VideoSeg;
                   break;

                 case ATT400MED:
                   TwoBitsPerPixel = FALSE;
                   GraphVadrRut = 1;
                   break;

                 default:
                   TwoBitsPerPixel = TRUE;
                   CurrScanBytes = 80;
                   GraphVadrRut = 1;
              }
              break;

        case HERCMONO:
              FontBitBlt = 1;
              TwoBitsPerPixel = FALSE;
              GraphVadrRut = 2;
              VideoSegPtr = HercVideoSeg;
              break;

        case PC3270:
              FontBitBlt = 1;
              CurrVideoSeg = 0xB800;
              TwoBitsPerPixel = FALSE;
              GraphVadrRut = 0;
              break;

        case VGA:
        case EGA:
              FontBitBlt = 2;
              CurrVideoSeg = 0xA000;
              break;

        default:
              FontBitBlt = 0;
    }
}

/*********************************************************************/
/*                                                                   */
/*                        loaduserfont                               */
/*                                                                   */
/*   Funkcija loaduserfont ucitava font iz Font Editor FNT datoteke  */
/*   i vraca njegov identifikator (handle) koji se dalje koristi     */
/*   pri ispisu teksta ili brisanju fonta (moze biti i nula).        */
/*   Ukoliko dodje do greske funkcija vraca negativan broj (isti     */
/*   kao i borlandove BGI funkcije tako da mozete koristiti          */
/*   BGI funkciju grapherrormsg( greska ); ). Ova funkcija ucitava   */
/*   samo fontove snimljene u screen formatu (pogledajte datoteku    */
/*   formati.txt)                                                    */
/*                                                                   */
/*   ULAZ: filename - ime datoteke na disku zajedno s direktorijem   */
/*                    ako je potrebno                                */
/*         facename - ime fonta (max 19 karaktera + terminator '\0') */
/*         pointsize - velicina fonta (ako je pointsize = 0 velicina */
/*                     fonta ce biti zanemarena tj. ucitat ce se     */
/*                     prvi font u datoteci s imenom facename        */
/*                                                                   */
/*   VRACA: font idetifikator (pozitivan broj ili nula) odnosno      */
/*          kod greske (negativan broj)                              */
/*                                                                   */
/*    -8  Datoteka ne postoji / ne moze se ucitati  grFontNotFound   */
/*    -12 Greska u datoteci (vjerovatno fizicka)    grIOerror        */
/*    -10 Nije user FNT datoteka                    grInvalidFont    */
/*    -11 Nije pronadjen trazeni font               grError          */
/*    -14 Instalirano vise od MAXUSERFONTS fontova  grInvalidFontNum */
/*    -9  Nedovoljno memorije za font               grNoFontMem      */
/*                                                                   */
/*********************************************************************/

int loaduserfont(const char *filename,  /* ime datot. u kojoj se nalazi font */
                 const char *facename,  /* ime fonta */
                 int        pointsize)  /* velicina fonta */
{
   FILE  *fp;

   char  buff[9];
   char  dskfacename[20];                /* ime fonta na disku */
   byte  dsksize;                        /* velicina fonta na disku */
   long  nextfont;                       /* pointer na slijedeci font */
   int   i,j,n;                          /* privremeni int */
   int   bmsize[256];                    /* duzine bitmapa */
   long  ubmsize;                        /* ukupna duzina svih bitmapa */
   struct fontheader *fonthead;          /* pointer na fonthead strukturu */
   byte  *fontwidth;
   byte  **bitmapadress;                 /* pointer na niz pointera */
   byte  *definedchar;
   int   hfont;                          /* handle na font */
   struct fontheader *fadress;


   /* potrazi slobodan handle za font */

   hfont = -14;                       /* grInvalidFontNum */
   for(i = 0; i < MAXUSERFONTS; i++)
      if(fontadress[i] == NULL) {
          hfont = i;
          break;
      }
   if(hfont < 0)
       return -14;                /* vec je instalirano MAXUSERFONTS */

   if((fp = fopen(filename, "rb")) == NULL)
         return -8;                  /* grFontNotFound */

   if(fread(buff, 5, 1, fp) != 1)
         return -12;                 /* grIOerror greska kod ucitavanja */

   if(strncmp(buff,"FEFNT",5) != 0)
         return -10;                 /* grInvalidFont nije user FNT datoteka */

   nextfont = 0;                     /* prvi font u datoteci je na ofsetu nula */

   do {
       fseek(fp, nextfont, SEEK_SET);     /* pozicioniranje na font */
       fread(buff, 8, 1, fp);             /* format fonta */
       fread(dskfacename, 20, 1, fp);     /* ime fonta na disku */
       dsksize = fgetc(fp);               /* velicina fonta */
       fread(&nextfont, 4, 1, fp);        /* adresa slijedeceg fonta u datoteci */

       if(strncmp(buff,"FEFNTSCR",8) != 0 || stricmp(dskfacename,facename) != 0
          || (dsksize != pointsize && pointsize))
       {
          /* ili nije screen format ili je razlicito ime fonta ili je
             razlicita visina (ako je pointsize=0 velicina se ne uzima
             u obzir ) */

          if(nextfont)          /* ako ovo nije zadnji font trazi dalje */
              continue;

          if(ferror(fp)) {
	     fclose(fp);
             return -12;       /* grIOerror */
          }
          fclose(fp);
          return -11;          /* grError   nije pronadjen trazeni font */
       }
       /* sve u redu font je pronadjen */
       /* alokacije memorije za strukturu, bitmape itd */

       if((fonthead = malloc(sizeof(struct fontheader))) == NULL)
       {
          fclose(fp);
          return -9;    /*  grNoFontMem */
       }

       strcpy(fonthead->name, dskfacename);
       fonthead->size = dsksize;
       fonthead->numchars = fgetc(fp);
       fonthead->defaultch = fgetc(fp);
       fonthead->height = fgetc(fp);
       fonthead->avwidth = fgetc(fp);
       fseek(fp, 2, SEEK_CUR);
       fonthead->baseline = fgetc(fp);

       if((fontwidth = malloc(256 * sizeof(unsigned char))) == NULL) {
            fclose(fp);
            free(fonthead);
            return -9;     /* grNoFontMem */
       }
       if((bitmapadress = malloc(256 * sizeof( byte *))) == NULL) {
            fclose(fp);
            free(fontwidth);
            free(fonthead);
            return -9;     /* grNoFontMem */
       }

       /* ovdje ce biti poredani svi definirani karakteri u fontu
          ( ne moraju biti po ASCII redosljedu */

       if((definedchar = malloc(fonthead->numchars * sizeof(unsigned char)))
	      == NULL)
       {
            fclose(fp);
            free(bitmapadress);
            free(fontwidth);
            free(fonthead);
            return -9;     /* grNoFontMem */
       }

       /* postavljamo sve sirine na 0 */
       memset(fontwidth, 0, 256*sizeof(byte));

       for(i = ubmsize = 0; i < fonthead->numchars; i++) {
          if((n = fgetc(fp)) != EOF) {
             fseek(fp, 1, SEEK_CUR);     /* preskacemo lijevu marginu */
             fontwidth[n] = fgetc(fp);   /* uzimamo sirinu */
             definedchar[i] = n;
             fseek(fp, 1, SEEK_CUR);
             bmsize[i] = getw(fp);
             ubmsize += bmsize[i];
	  }
          else
            fseek(fp, 5, SEEK_CUR);     /* u biti nepotrebno */
       }

       /* rezerviranje memorije i ucitavanje bitmapa karaktera */
       /* pokusajmo rezervirati memoriju u jednom bloku */

       if( ubmsize < 50000L &&
           (bitmapadress[definedchar[0]] = malloc(ubmsize)) != NULL)
       {
          fonthead->blockreserved = TRUE;   /* uspjelo je */
          for(i = 1; i < fonthead->numchars; i++)
             bitmapadress[definedchar[i]] = bitmapadress[definedchar[i-1]]
                                            + bmsize[i-1];
          fread(bitmapadress[definedchar[0]], ubmsize, 1, fp);
       }
       else {
          /* nemamo velik blok, probajmo ucitat font u djelovima */
          fonthead->blockreserved = FALSE;
	  for(i = 0; i < fonthead->numchars; i++) {
             if((bitmapadress[definedchar[i]] = malloc(bmsize[i])) == NULL)
             {
                 fclose(fp);
                 for(j = i - 1; j >= 0; j--)
                      free(bitmapadress[definedchar[j]]);
                 free(bitmapadress);
                 free(fontwidth);
                 free(fonthead);
                 free(definedchar);
                 return -9;             /* grNoFontMem */
             }
             fread(bitmapadress[definedchar[i]], bmsize[i], 1, fp);
          }
       }

       /* font je ucitan pogledajmo dali je doslo do I/O greske
          (fizicko ostecenje diska ili ostecena font datoteka) */

       if( ferror(fp))
       {
          fclose(fp);
          if(fonthead->blockreserved)
              free(bitmapadress[definedchar[0]]);
          else
              for(i = 0; i < fonthead->numchars; i++)
                  free(bitmapadress[definedchar[i]]);
          free(bitmapadress);
          free(fontwidth);
          free(fonthead);
          free(definedchar);
          return -12;             /* grIOerror */
       }
       /* postavljanje adrese i sirine default karaktera za sve one koji
          nisu definirani u fontu prepoznajemo ih po tome sto imaju
          sirinu 0 */

       fonthead->ptrwidth = fontwidth;
       fonthead->ptrbitmap = bitmapadress;
       fonthead->ptrdefined = definedchar;

       for(i = 0; i < 256; i++)
          if( ! fontwidth[i]) {
               fontwidth[i] = fontwidth[fonthead->defaultch];
               bitmapadress[i] = bitmapadress[fonthead->defaultch];
          }

       fontadress[hfont] = fonthead;
       return hfont;     /* font je instaliran vracamo handle na njega */

   } while(TRUE);
}



/*********************************************************************/
/*                                                                   */
/*                        deleteuserfont                             */
/*                                                                   */
/*   Funkcija deleteuserfont  brise Font Editor user font iz         */
/*   iz memorije i vraca vrijednost razlicitu od nule (TRUE)         */
/*   ako je funkcija uspjesno izvrsena odnosno false ako nije        */
/*   (font vjerovatno nije ni ucitan ili je zadan los handle)        */
/*                                                                   */
/*   ULAZ:  hfont - font idetifikator                                */
/*   IZLAZ: TRUE  - funkcija uspjesno izvrsena                       */
/*          FALSE - font nije ni ucitan ili pogresan handle          */
/*                                                                   */
/*********************************************************************/

int  deleteuserfont(int hfont)
{
   int i;
   struct fontheader *fnthead;


   if(hfont < 0 || hfont >= MAXUSERFONTS)
       return FALSE;

   /* fnthead smo definirali da se ne petljamo puno sa fontadress[hfont] */

   if((fnthead = fontadress[hfont]) == NULL)
       return FALSE;
   if(fnthead->blockreserved)
       free(fnthead->ptrbitmap[fnthead->ptrdefined[0]]);
   else
       for(i = 0; i < fnthead->numchars; i++)
           free(fnthead->ptrbitmap[fnthead->ptrdefined[i]]);
   free(fnthead->ptrbitmap);
   free(fnthead->ptrwidth);
   free(fnthead->ptrdefined);
   free(fnthead);
   fontadress[hfont] = NULL;
   return TRUE;
}


/*********************************************************************/
/*                                                                   */
/*                     deletealluserfonts                            */
/*                                                                   */
/*  Funkcija deletealluserfonts brise sve Font Editor user fontove   */
/*  iz memorije                                                      */
/*                                                                   */
/*********************************************************************/

void deletealluserfonts(void)
{
   int i;

   for(i = 0; i < MAXUSERFONTS; i++)
        deleteuserfont(i);
   selectedfont = 0;
}


/**********************************************************************/
/*                                                                    */
/*                       selectuserfont                               */
/*                                                                    */
/*  Ova funkcija selektira user font specificiran s hfont parametrom  */
/*  kao trenutno vazeci font za sve "user font" funkcije.             */
/*  U jednom trenutku mozete imati samo jedan selektiran font bez     */
/*  obzira koliko ste fontova ucitali u memoriju                      */
/*                                                                    */
/*  ULAZ:  hfont - idetifikator (handle) fonta                        */
/*  IZLAZ: funkcija vraca identifikator prije vazeceg fonta           */
/*                                                                    */
/*  NAPOMENA: Ukoliko hfont identifikator nije valjan stari font      */
/*            ostaje i dalje vazeci. Inicijalni identifikator je      */
/*            je identifikator prvog ucitanog fonta u memoriji.       */
/*            Ako se font koji je bio trenutno vazeci brise iz        */
/*            memorije funkcijom deleteuserfont obavezno pozovite ovu */
/*            funkciju (deletealluserfonts sama postavlja ovaj identi-*/
/*            fikator na inicijalnu vrijednost) . Mada je inicijalni  */
/*            identifikator zapravo identifikator prvog ucitanog      */
/*            fonta svejedno pozovite ovu funkciju i kod prvog        */
/*            ucitavanja.                                             */
/**********************************************************************/

int selectuserfont(int hfont)
{
   int  iret;

   iret = selectedfont;
   if(hfont >= 0 && hfont < MAXUSERFONTS && fontadress[hfont])
        selectedfont = hfont;
   return iret;
}


/********************************************************************/
/*                                                                  */
/*                     usertextwidth                                */
/*                                                                  */
/*  Vraca sirinu teksta u pikselima (kao BGI funkcija textwidth)    */
/*  0 ako user font nije instaliran                                 */
/********************************************************************/

int usertextwidth(const char *textstr)
{
   struct fontheader *fnthead;
   byte ch;
   int  width;

   width = 0;
   if((fnthead = fontadress[selectedfont]) != NULL)
       while((ch = *textstr++) != 0)
           width += fnthead->ptrwidth[ch];
   return width;
}


/********************************************************************/
/*                                                                  */
/*                     usertextheight                               */
/*                                                                  */
/*  Vraca ukupnu visinu teksta u pikselima (razlicito od BGI        */
/*  funkcije textheight koja ustvari vraca baznu liniju),           */
/*  0 ako user font nije instaliran                                 */
/********************************************************************/

int usertextheight(void)
{
   if(fontadress[selectedfont] != NULL)
      return fontadress[selectedfont]->height;
   else
      return 0;
}


/********************************************************************/
/*                                                                  */
/*                     userfontbaseline                             */
/*                                                                  */
/*  Vraca baznu liniju teksta u pikselima (isto kao i BGI funkcija  */
/*  textheight)                                                     */
/*  0 ako user font nije instaliran                                 */
/********************************************************************/

int userfontbaseline(void)
{
   if(fontadress[selectedfont] != NULL)
      return fontadress[selectedfont]->baseline;
   else
      return 0;
}


/********************************************************************/
/*                                                                  */
/*                    getuserfontheader                             */
/*                                                                  */
/*  Vraca pointer na fontheader strukturu fonta specificiranog sa   */
/*  hfont parametrom. Napominjem  da podatke iz ove strukture       */
/*  smijete samo citati. Bilo kakva promjena podataka u ovoj        */
/*  strukturi moze ugroziti zivot vaseg programa i samog MS DOS-a   */
/********************************************************************/

struct fontheader *getuserfontheader(int hfont)
{
    if(hfont < 0 || hfont >= MAXUSERFONTS)
        return NULL;
    return fontadress[hfont];
}


/****************************************************************/
/*                                                              */
/*                    setuserbkmode                             */
/*                                                              */
/* Ovom funkcijom se postavlja mod pozadine kod ispisivanja     */
/* teksta. Mod pozadine odredjuje da li ce pozadina biti ispu-  */
/* njena tekucom bojom pozadine (COPY_PUT) ili ce ostati nepro- */
/* mjenjena (OR_PUT). Ova funkcija ima istu svrhu kao i Windows */
/* API funkcija SetBkMode (COPY_PUT je OPAQUE, a OR_PUT ima     */
/* isti smisao kao TRANSPARENT). Inicijalna vrijednost je za    */
/* razliku od Windows-a OR_PUT, jer borlandove BGI funkcije za  */
/* ispis teksta podrzavaju samo taj nacin.Kod grafickih kartica */
/* za koje nisu napisane asemblerske rutine za ispis teksta     */
/* (IBM8514, EGA64, EGAMONO) tekst se ispisuje 3.5 puta brze u  */
/* OR_PUT nacinu. Funkcija vraca stari mod pozadine.            */
/*                                                              */
/* OR_PUT=2 i COPY_PUT=0 su definirani u graphic.h datoteci     */
/****************************************************************/

int setuserbkmode(int bkmode)
{
    int i;

    i = TextBkMode;
    TextBkMode = bkmode;
    return i;
}

/****************************************************************/
/*                                                              */
/*                    getuserbkmode                             */
/*                                                              */
/* Vraca mod pozadine za iscrtavanje karaktera.                 */
/****************************************************************/

int getuserbkmode(void)
{
    return TextBkMode;
}


/****************************************************************/
/*                                                              */
/*                       userouttext                            */
/*                                                              */
/* Ispisuje tekst na ekran kao i BGI funkcija outtext (tekst se */
/* ispisuje na poziciji grafickog kursora, koji se nakon ispisa */
/* pomice na kraj teksta). Od svih funkcija za ispis teksta ova */
/* je najsporija.                                               */
/****************************************************************/

int usertextout(char *str)
{
   int  x, y;
   struct textsettingstype fontinfo;

   x = getx();
   y = gety();
   gettextsettings(&fontinfo);
   if(fontinfo.direction == HORIZ_DIR && fontinfo.horiz == LEFT_TEXT)
   {
     moveto(x + usertextwidth(str), y);
   }
   return userTextOut(x, y, str, strlen(str));
}

/****************************************************************/
/*                                                              */
/*                       userouttextxy                          */
/*                                                              */
/*  ispisuje tekst na ekran potpuno jednako kao i BGI funkcija  */
/*  outtextxy (ne mijenja C.P.).                                */
/****************************************************************/

int userouttextxy(int x, int y, char *str)
{
   return userTextOut(x, y, str, strlen(str));
}

/*****************************************************************/
/*                                                               */
/*                       userTextOut                             */
/*                                                               */
/*  ispisuje tekst na ekran kao Windows API funkcija TextOut     */
/*                                                               */
/*****************************************************************/

/******************************************************************/
/*                                                                */
/*  Kako je brzina ispisa teksta u grafickom modu uvjek kriticna, */
/*  dijelovi ove funkcije (za graficke kartice VGA, EGA, PC3270,  */
/*  HERCULES, ATT400, MCGA i CGA) su napisane u asembleru. Ipak   */
/*  ni ovako napisana funkcija ne daje nimalo brz ispis, jer se   */
/*  slova "crtaju" tockicu po tockicu (bit po bit). Ispis mozete  */
/*  ubrzati 2-3 puta, ako slova "crtate" bajt po bajt (tj. ako u  */
/*  video memoriju upisujete do 8 tockica odjednom), kao sto je   */
/*  napisano za EGA/VGA karticu.  Trebalo bi                      */
/*  izbjeci izracunavanje video memorije za svako slovo, a bilo   */
/*  bi dobro za svaku graficku karticu posebno, funkciju napisati */
/*  kompletno u asembleru. Vrlo veliko povecanje brzine mozete    */
/*  dobiti ako napisete posebnu rutinu za svaki font (narucito    */
/*  ako je neproporcionalan).                                     */
/*                                                                */
/*  Uspredbe radi Windows ispisuje tekst oko 3 (~40% za VGA/EGA)  */
/*  puta brze od ove funkcije, BGI funkcija outtextxy (sa default */
/*  fontom) je jedva nesto brza kod mono i duplo sporija          */
/*  kod EGA/VGA. dok je ispis preko BIOS-a nesto manje od 2 puta  */
/*  brzi (otprilike isti za VGA/EGA), ali treba uzeti u obzir da  */
/*  BIOS ispisuje fiksni font na bajt poravnatoj adresi           */
/******************************************************************/

int userTextOut(int x0, int y0, char *str, int nCount)
{
   struct fontheader *fnthead;
   int    i, x, y, x1, wint;
   int    xmem, ymem;        /* stvarne koordinate u video memoriji */
   int    xmin, xmax;        /* min/max x koordinata */
   int    ybitmapoffset;     /* br.linija bitmape karakt. koje treba preskociti */
   int    height, width;
   /* int    chrsize;  */         /* fontinfo.size */
   unsigned BitMsegm;        /* segmentna adresa pbitmap */
   byte   *pwidth;
   byte   *pbitmap;
   struct textsettingstype fontinfo;
   struct viewporttype     viewport;
   byte   fgcolor, bkcolor;
   byte   mask,firstmask,lastmask,preskocilast;
   byte   ostatak,oldbitmap;
   int    xpixs, xbyte, wbyte;
   struct fillsettingstype fillinfo; /* za brisanje pozadine kod VERT_DIR && OR_PUT */

   byte   cgaparnepar;        /* kod CGAHI MCGAMED ATTMED itd parna neparna y linija */
   static byte monomask[12] = { 0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1, /* maske za 1 bit po pixelu */
				0xc0, 0x30, 0xc, 3 }; /* maske za 2 bita po pixelu */

   byte *pmonomask = monomask;        /* automatske inicijalizacije */
   byte  prTextbkmode = TextBkMode;   /* zbunjuju asm direktive */



   if(CurrGraphDriver == -1)
      return -1;                           /* grNoInitGraph */

   if((fnthead = fontadress[selectedfont]) == NULL)
      return -14;                          /* grInvalidFontNum */

   if((fgcolor = getcolor()) == (bkcolor = getbkcolor()))
	  return 0;

   height  = fnthead->height;
   pwidth  = fnthead->ptrwidth;
   getviewsettings(&viewport);
   gettextsettings(&fontinfo);
   /* chrsize = fontinfo.charsize; */

   if(fontinfo.direction == VERT_DIR) {
      /* koordinate lijeve gornje tocke prvog karaktra u stringu */
      if(fontinfo.vert != BOTTOM_TEXT) {
	 for(i = width = 0; i < nCount; i++)
	    width += fnthead->ptrwidth[str[i]];
	 if(fontinfo.vert == CENTER_TEXT)
	    y0 += width/2;
	 else
	    y0 += width;
      }
      switch(fontinfo.horiz) {

	/*    ************ NAPOMENA!!! **************

	   Borlandov BGI outtextxy (C++ 2.0) ima BUG kod VERT_DIR i LEFT_TEXT justify.
	   Ako zelite da ova funkcija tocno oponasa outtextxy uklonite oznake
	   komentara iz slijedece linije */
      /* case LEFT_TEXT: */
	 case RIGHT_TEXT: x0 -= height; break;
	 case CENTER_TEXT: x0 -= height / 2; break;
      }
   }
   else {  /* HORIZ_DIR */
      /* koordinate lijeve gornje tocke prvog karaktra u stringu */
      if(fontinfo.horiz != LEFT_TEXT) {
	 for(i = width = 0; i < nCount; i++)
	    width += fnthead->ptrwidth[str[i]];
	 if(fontinfo.horiz == CENTER_TEXT)
	    x0 -= width/2;
	 else
	    x0 -= width;
      }
      switch(fontinfo.vert) {
	 case BOTTOM_TEXT: y0 -= height; break;
	 case CENTER_TEXT: y0 -= height / 2; break;
      }

      xmem = x0 + viewport.left;  /* stvarne koordinate u video memoriji */
      ymem = y0 + viewport.top;
      ybitmapoffset = 0;          /* br linija bitmape kar. koje treba preskocit */
      if(viewport.clip) {
	if(y0 + height < viewport.top || y0 > viewport.bottom || x0 > viewport.right)
	   return 0;
	if(y0 + height > viewport.bottom)
	   height = viewport.bottom - y0 + 1;
	else
	   if(y0 < viewport.top) {
	      ybitmapoffset = viewport.top - y0;
	      height -= ybitmapoffset;
	      y0 = 0;
	      ymem = viewport.top;
	   }
	xmin = viewport.left;       /* min/max x pozicija */
	xmax = viewport.right;
      }
      else {
	if(ymem + height < 0 || ymem > CurrMaxY || xmem > CurrMaxX)
	   return 0;
	if(ymem + height > CurrMaxY)
           height = CurrMaxY - ymem + 1;
        else
           if(ymem < 0) {
              ybitmapoffset = -ymem;
              height -= ybitmapoffset;
              ymem = 0;
              y0 = -viewport.top;
           }
        xmin = 0;
        xmax = CurrMaxX;
      }
   }

   /* za default graficku karticu ili VERT_DIR i COPY_PUT obrisimo pozadinu */
   /* odjednom. Ovim se dobaja na brzini, ali moze doci do treperenja ako se */
   /* u petlji ispisuju tekstovi na istom mjestu (obicno neki brojaci, sat itd) */
   /* Ako vam ovo smeta izbrisite slijedecih 12 redova i uklonite oznake komentara */
   /* u rutini dvadesetak linija dolje */

   if((fontinfo.direction==VERT_DIR || FontBitBlt==0) && TextBkMode==COPY_PUT)
   {
      getfillsettings(&fillinfo);               /* sacuvajmo prijasnje stanje */
      for(i = width = 0; i < nCount; i++)
	 width += fnthead->ptrwidth[str[i]];    /* sirina teksta */
      setfillstyle(SOLID_FILL, bkcolor);        /* ispuni pravokutnik bojom pozadine */
      if(fontinfo.direction == VERT_DIR)
	 bar(x0, y0, x0 + height, y0 - width);
      else
	 bar(x0, y0, x0 + width, y0 + height);
      setfillstyle(fillinfo.pattern, fillinfo.color); /* vratimo prijasnju boju/srafuru */
   }

   for(i = 0; i < nCount; i++) {
      if((width = pwidth[(unsigned)*str]) == 0)
	   continue;
      pbitmap = fontadress[selectedfont]->ptrbitmap[(unsigned)*str++];
      if(fontinfo.direction == VERT_DIR || FontBitBlt == 0)
      {
	 /* nacrtajmo font spooooro */
	 wbyte = (width - 1)/8 + 1;
	 for(x = 0; x < width; x++)
	 {
	    xbyte = x / 8;
	    mask  = 0x80 >> x % 8;
	    xpixs = x0 + x;
	    for(y = 0; y < height; y++)
	    {
	       if( *(pbitmap + y * wbyte + xbyte) & mask)
		  if(fontinfo.direction == HORIZ_DIR)
		     putpixel(xpixs, y0 + y, fgcolor);
		  else
		     putpixel(x0 + y, y0 - x, fgcolor);
	       /*** ovdje ukloniti oznake komentara ***/
	       /* else
		  if(TextBkMode == COPY_PUT)
		     if(fontinfo.direction == HORIZ_DIR)
			putpixel(xpixs, y0 + y, bkcolor);
		     else
			putpixel(x0 + y, y0 - x, bkcolor);
	       */
	    }
	 }
	 if(fontinfo.direction == HORIZ_DIR)
	    x0 += width;
	 else
	    y0 -= width;
	 continue;  /* novi karakter */
      }
      else {
	 if(x0 >= xmin && x0+width <= xmax)
	 {
	     if(FontBitBlt == 1) /* CGA, MCGA, HERCULES, ATT400, PC3270 */
	     {
		   /* video adresa */
		      asm  mov   dx,CurrScanBytes
		      asm  mov   ax,ymem
		      asm  mov   es,CurrVideoSeg
		      asm  mov   cl,GraphVadrRut;  /* 0-MCGA 1-CGA 2-HERC */
		      asm  or    cl,cl
		      asm  jz    mcga1
		   /* CGA, MCGAMED, ATTMED itd */
		      asm  cmp   cl,2
		      asm  je    herc1
		      asm  xor   ch,ch
		      asm  shr   ax,1              /* za neparne linije + 0x2000 */
		      asm  pushf
		      asm  rcl   ch,1              /* 1 za nepar 0 za par */
		      asm  mov   cgaparnepar,ch
		      asm  mul   dx                /* CurrScanBytes * ymem/2 */
		      asm  popf
		      asm  jnb   adrend
		      asm  add   ax,0x2000
		      asm  jmp   adrend
		  /*  HERCULES i ATT400HI */
	      herc1:  asm  mov   cx,ax
		      asm  and   cx,3              /* cx=y%4 * 2 */
		      asm  mov   cgaparnepar,cl
		      asm  add   cx,cx
	      #if defined(__NEARPOINTERS)
		      asm  mov   bx,VideoSegPtr
		      asm  add   bx,cx
		      asm  mov   es,[bx]
	      #else
		      asm  les   bx,VideoSegPtr
		      asm  add   bx,cx
		      asm  mov   es,es:[bx]        /* cx= video segment */
	      #endif
		      asm  shr   ax,1              /* y/4 */
		      asm  shr   ax,1              /* dalje nastavlja ka mcga */
		   /* MCGAHI i PC3270 */
	      mcga1:  asm  imul  dx                /* CurrScanBytes * ymem */

	      adrend: asm  mov   dx, ax
		      asm  xor   bx, bx            /* bx = 0 */
		      asm  mov   ax, xmem
		      asm  mov   ch, TwoBitsPerPixel
		      asm  or    ch, ch            /* if(!TwoBitsPerPixel) */
		      asm  jz    mono1             /*    goto mono1 */
		      asm  add   ax, ax
	       mono1: asm  mov   bl, al
		      asm  and   bl, 7             /* bx = xmem % 8 */
		      asm  push  bx
		      asm  shr   ax, 1
		      asm  shr   ax, 1
		      asm  shr   ax, 1             /* ax = xmem / 8 */
		      asm  add   dx, ax            /* dx = video memorija */
		      asm  mov   ax, bx            /* ax = bx = xmem%8 */
		      asm  push  es
		#if defined(__NEARPOINTERS)
		      asm  mov   bx,pmonomask
		      asm  push  ds
		      asm  pop   es
		#else
		      asm  les   bx,pmonomask
		#endif
		      asm  or    ch,ch            /* if(!TwoBitsPerPixel) */
		      asm  jz    mono2             /*   goto mono2 */
		      asm  add   bx,8             /* uzima maske */
		      asm  shr   ax,1
	      mono2:  asm  add   bx,ax
		      asm  mov   ah, es:[bx]       /* ah = maska */
		      asm  pop   es
		      asm  pop   bx
		      asm  or    ch,ch            /* if(!TwoBitsPerPixel) */
		      asm  jz    mono3
		      asm  mov   cl,6             /* bl="adresa" pixela u bajtu */
		      asm  sub   cl,bl            /* za cl ce trebati << boje */
		      asm  mov   bl,bkcolor
		      asm  mov   bh,fgcolor
		      asm  shl   bl,cl
		      asm  shl   bh,cl
		      asm  jmp   mono4
	      mono3:  asm  xor   bl, bl
		      asm  mov   bh, ah            /* bkcolor */
		      asm  cmp   BYTE PTR bkcolor, 0
		      asm  je    mono4
		      asm  xchg  bl, bh            /* fgcolor=0, bkcolor = 1 */
	      mono4:  asm  not   ah
		      asm  mov   al,prTextbkmode  /* COPY_PUT=0 ili OR_PUT */
		      asm  mov   di,dx          /* video adresa */
		      asm  push  ax
		      asm  mov   ax, ybitmapoffset
		      asm  mov   dx, WORD PTR pbitmap
		      asm  add   ax, dx
		 #if defined(__FARPOINTERS)
		      asm  mov   dx, WORD PTR pbitmap+2
		 #elif defined(__NEARPOINTERS)
		      asm  mov   dx,ds
		 #else
		 #error Unsupported memory model
		 #endif
		      asm  mov   BitMsegm,dx      /* seg adr bitmapa */
		      asm  mov   si, ax

		      asm  mov   ax, height
		      asm  mov   y, ax         /* for y = 0; y < height */
		      asm  pop   ax
	      loop0y: /* asm  mov   ch, BYTE PTR chrsize   povecanje po y */

		      /* ah = maska, al=PUT_COPY ili PUT_OR ch= charsize */
		      /* bh=fgcolor, bl=bkcolor, es=video segment */
		      /* di=video adresa, si=pointer na bitmapu */

	      loop1y: asm  push  ax           /* for y=0; y<height;y++ */
		      asm  push  bx
		      asm  push  di
		      asm  mov   dx, width
		      asm  mov   x, dx
		      asm  mov   xbyte, 8
	      loop1x: asm  mov   dh, BYTE PTR xbyte
		      asm  and   dh, 7        /* ako je djljiv sa 8 */
		      asm  jnz   mono8        /* treba uzet novi bajt */
		      asm  mov   xbyte,8
		      asm  mov   dx, BitMsegm /* bitmape */
		      asm  push  es
		      asm  mov   es, dx
		      asm  mov   dl, es:[si]
		      asm  inc   si
		      asm  pop   es
	      mono8:  /* asm  mov   cl, BYTE PTR chrsize   povecanje po x */

	     putpx1:  asm  mov   dh, es:[di]  /* video memorija */
		      asm  and   dh, ah       /* izdvoji piksel */
		      asm  test  dl, 0x80     /* dali je crna tocka */
		      asm  jne   crna1
		      asm  or    al,al
		      asm  jnz   put1or
		      asm  or    dh, bl       /* boja pozadine */
		      asm  jmp   put1cpy
	      crna1:  asm  or    dh, bh       /* fgcolor */
	     put1cpy: asm  mov   es:[di], dh
	      put1or: asm  ror   bl,1
		      asm  ror   bh,1
		      asm  ror   ah,1
		      asm  cmp   BYTE PTR TwoBitsPerPixel, 0
		      asm  je    no2bits
		      asm  ror   bl,1
		      asm  ror   bh,1
		      asm  ror   ah,1
	     no2bits: asm  and   ah,ah
		      asm  js    incmem
		      asm  inc   di          /* slijed byt vid mem */
	     incmem:  /* asm  dec   cl */
		      /* asm  jne   putpx1 */

		      asm  rol   dl,1        /* slijedeca tocka */
		      asm  dec   BYTE PTR xbyte
		      asm  dec   BYTE PTR x
		      asm  jne   loop1x

		      /* video adresa slijedece linije */
	    nline1:   asm  pop   di                  /* video adresa */
		      asm  mov   al,GraphVadrRut
		      asm  or    al,al
		      asm  jz    mcga2
		      asm  cmp   al,2
		      asm  je    herc2
		      /* CGA i MCGAMED, ATT400MED itd */
		      asm  mov   al,cgaparnepar
		      asm  xor   al,1                /* par ili nepar */
		      asm  mov   cgaparnepar,al
		      asm  test  al,1
		      asm  jz    cgapar
		      asm  add   di,0x2000
		      asm  jmp   nlinend
	     cgapar:  asm  sub   di,0x1fb0
		      asm  jmp   nlinend
		      /* HERCULES, ATT400HI */
	     herc2:   asm  mov   al,cgaparnepar  /* 0,1,2 ili 3 */
		    /*  asm  push  cx (za velicinu) */
		      asm  mov   cl,al
		      asm  inc   cl
		      asm  and   cl,3
		      asm  jne   herc4dl
		      asm  add   di,CurrScanBytes  /* linija je / sa 4 */
	     herc4dl: asm  and   al,0xfc
		      asm  or    al,cl          /* if((++cgaparnepar)==4) cgaparnepar=0 */
		      asm  mov   cgaparnepar,cl
		      asm  add   cl,cl
		      asm  xor   ch,ch
		  #if defined(__NEARPOINTERS)
		      asm  mov   bx,VideoSegPtr
		      asm  add   bx,cx
		      asm  mov   es,[bx]
		  #else
		      asm  les   bx,VideoSegPtr
		      asm  add   bx,cx
		      asm  mov   es,es:[bx]
		  #endif
		   /* asm pop cx (za velicinu) */
		      asm  jmp   nlinend
		      /* MCGA, PC3270 */
	      mcga2:  asm  add   di,CurrScanBytes   /* slijedece linije */

	     nlinend: asm  pop   bx
		      asm  pop   ax
		    /*  asm  dec   ch */
		    /*  asm  jne   loop1y */
		      asm  dec   BYTE PTR y
		      asm  je    monoend
		      asm  jmp   loop0y       /* skok veci od 127 bajtova */

	     monoend: asm  nop

	     }
	     else  /************** EGA / VGA ************************/
	     {

		   /* outi (write mode) */
		      asm  mov   dx,0x3ce
		      asm  mov   al,5
		      asm  out   dx,al
		      asm  inc   dx
		      asm  mov   al,2         /* EGA/VGA write mode 2 */
		      asm  out   dx,al
		      asm  dec   dx
		      asm  mov   al,8         /* EGA/VGA bit mask */
		      asm  out   dx,al

		   /* video adresa */
		      asm  mov   ax,ymem
		      asm  add   ax,ax        /* ymem * 2 */
		      asm  mov   bx,ax
		      asm  mov   cl,2
		      asm  shl   ax,cl        /* ax = ymem * 8 */
		      asm  add   ax,bx        /* ax = ymem * 10 */
		      asm  inc   cl
		      asm  shl   ax,cl        /* ax = ymem * 80 (CurrScanBytes) */
		      asm  mov   dx,xmem
		      asm  mov   bx,dx
		      asm  shr   dx,cl        /* dx = xmem/8 */
		      asm  add   dx,ax        /* dx = ofset video mem */

		      /* maske */
		      asm  mov   cx,bx
		      asm  and   cx,7         /* ostatak = xmem%8 */
		      asm  mov   ostatak,cl
		      asm  mov   ax,width     /* xbyte=(width-1+ostatak)/8+1 */
		      asm  add   ax,cx
		      asm  dec   ax
		      asm  shr   ax,1
		      asm  shr   ax,1
		      asm  shr   ax,1
		      asm  inc   ax
		      asm  mov   xbyte,ax
		      asm  cmp   BYTE PTR prTextbkmode,COPY_PUT
		      asm  jne   eganomsk     /* za OR_PUT nam maske nisu potrebne */
		      asm  mov   al,0xff      /* firstmask = 0xff >> ostatak */
		      asm  shr   al,cl
		      asm  mov   firstmask,ah
		      asm  mov   ax,width     /* lastmask = 0xff << (8-((width+xmem) & 7) */
		      asm  add   ax,xmem
		      asm  and   al,7
		      asm  mov   cl,8
		      asm  sub   cl,al
		      asm  mov   al,0xff
		      asm  shl   al,cl
		      asm  mov   lastmask,al
	    eganomsk: asm  mov   ax,width     /* if((width-1)/8+1 != xbytes) */
		      asm  dec   ax           /*   preskocilast=TRUE */
		      asm  shr   ax,1         /* else */
		      asm  shr   ax,1         /*   preskocilast=FALSE */
		      asm  shr   ax,1
		      asm  inc   ax
		      asm  sub   ax,xbyte     /* 0=FALSE 0xff=TRUE*/
		      asm  mov   preskocilast,al
		      asm  mov   es,CurrVideoSeg
		      asm  mov   bl,bkcolor
		      asm  mov   bh,fgcolor
		      asm  mov   di,dx          /* video adresa */
		      asm  mov   si, ybitmapoffset
		      asm  mov   dx, WORD PTR pbitmap
		      asm  add   si, dx
		 #if defined(__FARPOINTERS)
		      asm  mov   dx, WORD PTR pbitmap+2
		 #elif defined(__NEARPOINTERS)
		      asm  mov   dx,ds
		 #else
		 #error Unsupported memory model
		 #endif
		      asm  mov   BitMsegm,dx      /* seg adr bitmapa */

		      /* ah=ostatak */
		      /* bh=fgcolor, bl=bkcolor, es=video segment */
		      /* di=video adresa, si=pointer na bitmapu */
		      /* dx=3cf EGA/VGA out port (postavljanje maske) */

		      asm  mov   ah, ostatak
		      asm  mov   dx, height
		      asm  mov   y, dx         /* for y = 0; y < height; */
		      asm  mov   dx,0x3cf      /* out port */

	     eloop1y: asm  push  di           /* for y=height;y>0;y-- */
		      asm  mov   oldbitmap, 0
		      /* for x=xbyte; x>0; x-- (x je vec postavljen na xbyte) */
		      asm  mov   al,BYTE PTR xbyte
		      asm  mov   BYTE PTR x,al
	     eloop1x: asm  mov   ch,oldbitmap
		      asm  mov   cl,8
		      asm  sub   cl,ah
		      asm  shl   ch,cl     /* ch = oldbitmap << 8-ostatak */
		      asm  cmp   BYTE PTR x,1
		      asm  jne   enopres     /* if x!=1 || preskocilast == FALSE */
		      asm  cmp   BYTE PTR preskocilast,0
		      asm  je    enopres
		      asm  mov   al,ch
		      asm  jmp   preskok
	     enopres: asm  push  es
		      asm  mov   es, BitMsegm
		      asm  mov   al, es:[si]   /* bajt bitmape */
		      asm  mov   oldbitmap,al
		      asm  inc   si
		      asm  pop   es
		      asm  mov   cl,ah
		      asm  shr   al,cl     /* bitmap = ch | (*pbitmap++ >> ostatak) */
		      asm  or    al,ch

	     preskok: asm  out   dx,al        /* dx=3cf al=maska */
		      asm  mov   ch,bh
		      asm  xchg  es:[di],ch   /* boja slova (fgcolor) */
		      asm  cmp   BYTE PTR prTextbkmode,COPY_PUT
		      asm  jne   eput1or      /* skok ako je OR_PUT */
		      asm  mov   ch,0xff      /* bkmask=0xff */
		      asm  mov   cl,BYTE PTR x      /* if x==xbyte */
		      asm  cmp   cl,BYTE PTR xbyte
		      asm  jne   enofirst     /* skok ako nije prva iteracija */
		      asm  mov   ch,firstmask
	    enofirst: asm  cmp   cl,1         /* if x==1 */
		      asm  jne   enolast      /* skok ako nije zadnja iteracija */
		      asm  and   ch,lastmask
	    enolast:  asm  not   al           /* mask = ~mask && bkmask */
		      asm  and   al,ch
		      asm  out   dx,al
		      asm  mov   ch,bl
		      asm  xchg   es:[di],ch   /* boja pozadine */
	    eput1or:  asm  inc   di           /* slijedeci bajt video memorije */
		      asm  dec   BYTE PTR x
		      asm  jne   eloop1x      /* next x */

		      /* video adresa slijedece linije */
		      asm  pop   di                  /* video adresa */
		      asm  add   di,CurrScanBytes   /* slijedece linije */
		      asm  dec   BYTE PTR y
		      asm  jne   eloop1y

		      asm  mov   dx,0x3ce
		      asm  mov   ax,5          /* RAM addresing */
		      asm  out   dx,ax
		      asm  mov   ax,3          /* rotate / no change write mode */
		      asm  out   dx,ax
		      asm  mov   ax,0xff08     /* bit mask */
		      asm  out   dx,ax
		      asm  mov   ax,0xf02      /* character map mask */
		      asm  mov   dx,0x3c4
		      asm  out   dx,ax

	     }

	 }
	 x0 += width;
	 xmem += width;
      }
   }

   return 0;
}




/********************* PROBA ************************/


#include <time.h>

int main(void)
{
   /* request auto detection */
   int  gdriver = DETECT, gmode, errorcode;
   int  midx, midy, i;
   int  hfont;
   int  size;
   char buff[200];
   clock_t start,end;

   /* initialize graphics and local variables */
   initgraph(&gdriver, &gmode, "C:\\BORLANDC\\BGI");
   userinitgraph(gdriver, gmode);

   /* read result of initialization */
   errorcode = graphresult();
   if (errorcode != grOk)  /* an error occurred */
   {
      printf("Graphics error: %s\n", grapherrormsg(errorcode));
      printf("Press any key to halt:");
      getch();
      exit(1); /* terminate with an error code */
   }

   midx = getmaxx() / 2;
   midy = getmaxy() / 2;
   hfont = loaduserfont("FEFNT.FNT","System",10);
   if(hfont < 0) {
      closegraph();
      printf("Font error: %s\n",grapherrormsg(hfont));
      getch();
      exit(1);
   }
   selectuserfont(hfont);
   outtextxy(midx,0,"start");
   settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
   settextjustify(LEFT_TEXT, TOP_TEXT);
   setuserbkmode(OR_PUT);
   strcpy(buff,"Testiranje: Windows sistemski font.");
   
   start = clock();
   for( i = 0; i < 100; i++)
       userouttextxy(midx, midy, buff);
   end = clock();
   
   sprintf(buff,"time:%f   char per sec:%f",(end-start)/CLK_TCK, (100 * strlen(buff))/((end-start)/CLK_TCK));
   outtextxy(midx/2,midy/2,buff);

   deletealluserfonts();

   /* clean up */
   getch();
   closegraph();
   return 0;
}
